/*******************************************************************************
*                                                                              *
* foo_dsp_stereoconv                                                           *
* Copyright (C) 2007  Eric Larson (larson.eric.d@gmail.com)                    *
*                                                                              *
* This program is free software; you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation; either version 2 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    *
*                                                                              *
*******************************************************************************/

typedef signed __int64		sint64;
typedef unsigned __int64	uint64;
typedef signed int			sint32;
typedef unsigned int		uint32;
typedef signed short		sint16;
typedef unsigned short		uint16;
typedef signed char			sint8;
typedef unsigned char		uint8;

typedef sint64				int64;
typedef sint32				int32;
typedef sint16				int16;
typedef sint8				int8;

#include <math.h>
#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"
#include "resource.h"

#define TWOPI 6.283185307179586476925286766559
#define INVSQRT2 0.70710678118654752440084436210485
#define NC 1024
#define ML 1024

unsigned GetWindowLen(unsigned llen, unsigned rlen)
{
	unsigned out = 1;
	unsigned loggie = (int)ceil(log((double)max(max(llen,rlen),ML))/log(2.0));
	for(unsigned i=0; i<loggie; i++)
			out *= 2;
	return out;
}



bool get_wav_2ch(char const* p_path, unsigned* length, short* chs, double* ldata, double* rdata, abort_callback& p_abort)
{
    try
    {
		file_info_impl info;
        audio_chunk_impl chunk;
        input_helper helper;
		*length = 0;

        // open input
        helper.open(service_ptr_t<file>(), make_playable_location(p_path, 0), input_flag_simpledecode, p_abort);

        helper.get_info(0, info, p_abort);
        if (info.get_length() <= 0) return false;
        if (!helper.run(chunk, p_abort)) return false;
		if (chunk.get_channels() > 2) return false;
		else if(chs) (*chs) = chunk.get_channels();
		(*length) = (unsigned)audio_math::time_to_samples(info.get_length(), chunk.get_sample_rate());

		if(ldata != NULL) {
			// Store the data somewhere.
			if (chunk.get_channels() == 2) {
				while (true)
				{
					const audio_sample *inSamples = chunk.get_data();
					int sampleCount = (int)chunk.get_sample_count();

					for (int i = 0; i < sampleCount; i++) {
						ldata[0] = (double)inSamples[0];
						rdata[0] = (double)inSamples[1];
						ldata++;
						rdata++;
						inSamples += 2;
					}

					if (!helper.run(chunk, p_abort)) break;
				}
			}
			else { // 1 channel
				while (true)
				{
					const audio_sample *inSamples = chunk.get_data();
					int sampleCount = (int)chunk.get_sample_count();

					for (int i = 0; i < sampleCount; i++) {
						ldata[0] = (double)inSamples[0];
						ldata++;
						inSamples ++;
					}

					if (!helper.run(chunk, p_abort)) break;
				}
			}
		}

        return true;
    }
    catch (const exception_aborted &)  { return false; }
    catch (const std::exception & exc) { return false; }
}

struct t_dsp_stereoconv_params
{
	double leveldB;
	char lFile[NC];
	char rFile[NC];

	// The constructor sets the default values.
	t_dsp_stereoconv_params() { leveldB = -4.5; strcpy_s(lFile,NC,"l.wav"); strcpy_s(rFile,NC,"r.wav"); }

	// Read in data from a preset.
	bool set_data(const dsp_preset & p_data)
	{
		if (p_data.get_data_size() != sizeof(t_dsp_stereoconv_params)) return false;
		t_dsp_stereoconv_params temp = *(t_dsp_stereoconv_params *)p_data.get_data();
		byte_order::order_le_to_native_t(temp);
		leveldB = temp.leveldB; // Set it here (get it from the passed arg)
		strcpy_s(lFile,NC,temp.lFile);
		strcpy_s(rFile,NC,temp.rFile);
		return true;
	}

	// Write data to a preset.
	bool get_data(dsp_preset & p_data)
	{
		t_dsp_stereoconv_params temp;
		temp.leveldB = leveldB;
		strcpy_s(temp.lFile,NC,lFile);
		strcpy_s(temp.rFile,NC,rFile);
		byte_order::order_native_to_le_t(temp);
		p_data.set_data(&temp, sizeof(temp)); // Get it here (set it in the passed arg)
		return true;
	}
};

class dsp_abortback : public abort_callback_impl
{
public:
	dsp_abortback() { }
	bool is_aborting() { return false; }
	abort_callback_event get_abort_event() { return 0; }
};

class Convolve {
public:
	Convolve(t_dsp_stereoconv_params & Params) {
		Start(Params);
	}

	~Convolve() {
		Finish();
	}

	bool ProcessSamples(int sampleCount) {
		mInputPos = (mInputPos + sampleCount) & (kWindowSize - 1);
		mInputSamplesNeeded -= sampleCount;

		if (mInputSamplesNeeded == 0) {
			Run();
			return true;
		}

		return false;
	}

	int FlushWithReturn() {
		double (*zeroBuff)[2] = new double[kWindowSize][2];
		int returnCount;

		memset(zeroBuff, 0, sizeof(zeroBuff[0][0])*kWindowSize*2);

		if (!mFlushing) {
			mFlushing = true;
			mFlushSamplesRemaining = kWindowSize - mInputSamplesNeeded;
		}
		else if (mFlushSamplesRemaining <= 0) {
			Flush();
			return -1;
		}

		memcpy(&mInput[mInputPos][0], &zeroBuff[0][0], mInputSamplesNeeded * 2 * sizeof(double));
		if (ProcessSamples(mInputSamplesNeeded)) {
			returnCount = min(mFlushSamplesRemaining, (int)kWindowSize);
		}
		else {
			returnCount = 0;
		}
		mFlushSamplesRemaining -= mInputSamplesNeeded;
		delete [] zeroBuff;
		return returnCount;
	}

	void Flush() {
		mFlushing = false;
		mInputPos = 0;
		mInputSamplesNeeded = kWindowSize;
	}

	double *GetInputBuffer() {
		return &mInput[mInputPos][0];
	}

	int GetInputSamplesNeeded() {
		return mInputSamplesNeeded;
	}

	double *GetOutputBuffer() {
		return &mOutput[0][0];
	}

	int GetOutputLength() {
		return kWindowSize;
	}

private:
	unsigned		kWindowSize;
	unsigned		kTransSize;
	unsigned		e;
	double			SCALE;
	unsigned		RESETSIZE;
	double			mLeveldB;
	bool			mFlushing;
	int				mFlushSamplesRemaining;
	uint32			mInputSamplesNeeded;
	uint32			mInputPos;
	unsigned		(*mBitRev);
	unsigned		(*mBitFwd);
	char			lFile[NC];
	char			rFile[NC];
	double			(*mSineTab);
	double			(*mInput)[2];
	double			(*mOutput)[2];
	double			(*mOldLTail);
	double			(*mOldRTail);
	double			(*mTempLBuffer);
	double			(*mTempRBuffer);
	double			(*mOutLBuffer);
	double			(*mOutRBuffer);
	double			(*mFilterLtoLs);
	double			(*mFilterLtoLd);
	double			(*mFilterLtoRs);
	double			(*mFilterLtoRd);
	double			(*mFilterRtoLs);
	double			(*mFilterRtoLd);
	double			(*mFilterRtoRs);
	double			(*mFilterRtoRd);

	unsigned IntegerLog2(unsigned v) {
		unsigned i = 0;

		while(v>1) {
			++i;
			v >>= 1;
		}

		return i;
	}

	unsigned RevBits(unsigned x, unsigned bits) {
		unsigned y = 0;

		while(bits--) {
			y = (y+y) + (x&1);
			x >>= 1;
		}

		return y;
	}

	void VDCreateHalfSineTable(double *dst, int n) {
		const double twopi_over_n = TWOPI / n;

		for(int i=0; i<n; ++i) {
			dst[i] = sin(twopi_over_n * i);
		}
	}

	void VDCreateBitRevTable(unsigned *dst, int n) {
		unsigned bits = IntegerLog2(n);

		for(int i=0; i<n; ++i) {
			dst[i] = RevBits(i, bits);
		}
	}

	void VDCreateBitFwdTable(unsigned *fwd, const unsigned *rev, unsigned n) {
		for(unsigned i=0; i<n; ++i) {
			fwd[i] = rev[i] & (n-1);
		}
	}

	void VDComputeFHT(double *A, int nPoints, const double *sinTab) {
		int i, n, n2, theta_inc;

		// FHT - stage 1 and 2 (2 and 4 points)

		for(i=0; i<nPoints; i+=4) {
			const double	x0 = A[i];
			const double	x1 = A[i+1];
			const double	x2 = A[i+2];
			const double	x3 = A[i+3];

			const double	y0 = x0+x1;
			const double	y1 = x0-x1;
			const double	y2 = x2+x3;
			const double	y3 = x2-x3;

			A[i]	= y0 + y2;
			A[i+2]	= y0 - y2;

			A[i+1]	= y1 + y3;
			A[i+3]	= y1 - y3;
		}

		// FHT - stage 3 (8 points)

		for(i=0; i<nPoints; i+= 8) {
			double alpha, beta;

			alpha	= A[i+0];
			beta	= A[i+4];

			A[i+0]	= alpha + beta;
			A[i+4]	= alpha - beta;

			alpha	= A[i+2];
			beta	= A[i+6];

			A[i+2]	= alpha + beta;
			A[i+6]	= alpha - beta;

			alpha	= A[i+1];

			const double beta1 = INVSQRT2*(A[i+5] + A[i+7]);
			const double beta2 = INVSQRT2*(A[i+5] - A[i+7]);

			A[i+1]	= alpha + beta1;
			A[i+5]	= alpha - beta1;

			alpha	= A[i+3];

			A[i+3]	= alpha + beta2;
			A[i+7]	= alpha - beta2;
		}

		n = 16;
		n2 = 8;
		theta_inc = nPoints >> 4;

		while(n <= nPoints) {
			for(i=0; i<nPoints; i+=n) {
				int j;
				int theta = theta_inc;
				double alpha, beta;
				const int n4 = n2>>1;

				alpha	= A[i];
				beta	= A[i+n2];

				A[i]	= alpha + beta;
				A[i+n2]	= alpha - beta;

				alpha	= A[i+n4];
				beta	= A[i+n2+n4];

				A[i+n4]		= alpha + beta;
				A[i+n2+n4]	= alpha - beta;

				for(j=1; j<n4; j++) {
					double	sinval	= sinTab[theta];
					double	cosval	= sinTab[theta + (nPoints>>2)];

					double	alpha1	= A[i+j];
					double	alpha2	= A[i-j+n2];
					double	beta1	= A[i+j+n2]*cosval + A[i-j+n ]*sinval;
					double	beta2	= A[i+j+n2]*sinval - A[i-j+n ]*cosval;

					theta	+= theta_inc;

					A[i+j]		= alpha1 + beta1;
					A[i+j+n2]	= alpha1 - beta1;
					A[i-j+n2]	= alpha2 + beta2;
					A[i-j+n]	= alpha2 - beta2;
				}
			}

			n *= 2;
			n2 *= 2;
			theta_inc >>= 1;
		}
	}


	bool Start(t_dsp_stereoconv_params & Params) {
		unsigned i;
		mFlushing = false;

		strcpy_s(lFile,NC,Params.lFile);
		strcpy_s(rFile,NC,Params.rFile);
		mLeveldB = Params.leveldB;

		// Set the window and transform size based on IR length
		dsp_abortback abort_cback;
		unsigned llen = 0;
		unsigned rlen = 0;
		short lchs = 0;
		short rchs = 0;
		get_wav_2ch(lFile,&llen,&lchs,NULL,NULL,abort_cback);
		get_wav_2ch(rFile,&rlen,&rchs,NULL,NULL,abort_cback);

		kWindowSize = GetWindowLen(llen,rlen);
		kTransSize = 2*kWindowSize;
		SCALE = pow(10,mLeveldB/20.0)/(double)kWindowSize/4;
		RESETSIZE = (sizeof(double))*(kTransSize);

		// Read in audio data
		double* lldata = new double[kTransSize];
		double* lrdata = new double[kTransSize];
		double* rldata = new double[kTransSize];
		double* rrdata = new double[kTransSize];
		memset(lldata, 0, RESETSIZE);
		memset(lrdata, 0, RESETSIZE);
		memset(rldata, 0, RESETSIZE);
		memset(rrdata, 0, RESETSIZE);

		if(llen > 0) // Doesn't matter how many chs b/c goes into L
			get_wav_2ch(lFile,&llen,&lchs,lldata,lrdata,abort_cback);
		else
			lldata[0] = 1; // L->L is an impulse
		if(rlen > 0)
			if(rchs == 2)
				get_wav_2ch(rFile,&rlen,&rchs,rldata,rrdata,abort_cback);
			else // If one channel, must go into right input
				get_wav_2ch(rFile,&rlen,&rchs,rrdata,rldata,abort_cback);
		else
			rrdata[0] = 1; // R->R is an impulse

		// Allocate necessary memory
		mInput = new double[kWindowSize][2];
		mOutput = new double[kWindowSize][2];
		mOldLTail = new double[kWindowSize];
		mOldRTail = new double[kWindowSize];
		mBitRev = new unsigned[kTransSize];
		mBitFwd = new unsigned[kTransSize];
		mSineTab = new double[kTransSize];
		mTempLBuffer = new double[kTransSize];
		mTempRBuffer = new double[kTransSize];
		mOutLBuffer = new double[kTransSize];
		mOutRBuffer = new double[kTransSize];
		mFilterLtoLs = new double[kWindowSize];
		mFilterLtoLd = new double[kWindowSize];
		mFilterLtoRs = new double[kWindowSize];
		mFilterLtoRd = new double[kWindowSize];
		mFilterRtoLs = new double[kWindowSize];
		mFilterRtoLd = new double[kWindowSize];
		mFilterRtoRs = new double[kWindowSize];
		mFilterRtoRd = new double[kWindowSize];

		VDCreateBitRevTable(mBitRev, kTransSize);
		VDCreateBitFwdTable(mBitFwd, mBitRev, kTransSize);
		VDCreateHalfSineTable(mSineTab, kTransSize);

		mInputSamplesNeeded = kWindowSize;
		mInputPos = 0;

		memset(mInput, 0, (sizeof mInput[0][0])*kTransSize);
		memset(mOldLTail, 0, sizeof(double)*kWindowSize);
		memset(mOldRTail, 0, sizeof(double)*kWindowSize);

		// Reverse Bits for Henley Transform
		// Load into mTempLBuffer and such to conserve memory

		for(i=0; i<kTransSize; ++i){
			mTempLBuffer[i] = lldata[i];
			mTempRBuffer[i] = lrdata[i];
			mOutLBuffer[i] = rldata[i];
			mOutRBuffer[i] = rrdata[i];
		}
		for(i=0; i<kTransSize; ++i) {
			lldata[i] = mTempLBuffer[mBitFwd[i]]; // L to L
			lrdata[i] = mTempRBuffer[mBitFwd[i]]; // L to R
			rldata[i] = mOutLBuffer[mBitFwd[i]];  // R to L
			rrdata[i] = mOutRBuffer[mBitFwd[i]];  // R to R
		}

		VDComputeFHT(lldata, kTransSize, mSineTab); // L to L
		VDComputeFHT(lrdata, kTransSize, mSineTab); // L to R
		VDComputeFHT(rldata, kTransSize, mSineTab); // R to L
		VDComputeFHT(rrdata, kTransSize, mSineTab); // R to R

		mFilterLtoLs[0] = 2*lldata[0];
		mFilterLtoRs[0] = 2*lrdata[0];
		mFilterRtoLs[0] = 2*rldata[0];
		mFilterRtoRs[0] = 2*rrdata[0];
		mFilterLtoLd[0] = 0;
		mFilterLtoRd[0] = 0;
		mFilterRtoLd[0] = 0;
		mFilterRtoRd[0] = 0;

		for(i=1; i<kWindowSize; ++i)
		{
			const unsigned ni = kTransSize-i;
			mFilterLtoLs[i] = lldata[i] + lldata[ni];
			mFilterLtoLd[i] = lldata[i] - lldata[ni];
			mFilterLtoRs[i] = lrdata[i] + lrdata[ni];
			mFilterLtoRd[i] = lrdata[i] - lrdata[ni];
			mFilterRtoLs[i] = rldata[i] + rldata[ni];
			mFilterRtoLd[i] = rldata[i] - rldata[ni];
			mFilterRtoRs[i] = rrdata[i] + rrdata[ni];
			mFilterRtoRd[i] = rrdata[i] - rrdata[ni];
		}

		delete [] lldata;
		delete [] lrdata;
		delete [] rldata;
		delete [] rrdata;

		return true;
	}

	void Finish() {
		delete [] mInput;
		delete [] mOutput;
		delete [] mOldLTail;
		delete [] mOldRTail;
		delete [] mBitRev;
		delete [] mSineTab;
		delete [] mTempLBuffer;
		delete [] mTempRBuffer;
		delete [] mOutLBuffer;
		delete [] mOutRBuffer;
		delete [] mFilterLtoLs;
		delete [] mFilterLtoLd;
		delete [] mFilterLtoRs;
		delete [] mFilterLtoRd;
		delete [] mFilterRtoLs;
		delete [] mFilterRtoLd;
		delete [] mFilterRtoRs;
		delete [] mFilterRtoRd;
	}

	bool Run() {
		unsigned i, ni;

		// Create temporary buffer

		memset(mOutLBuffer, 0, RESETSIZE);
		memset(mOutRBuffer, 0, RESETSIZE);
		for(i=0; i<kWindowSize; ++i) {
			mOutLBuffer[i] = mInput[i][0];
			mOutRBuffer[i] = mInput[i][1];
		}

		// Reverse Bits for Hartley Transform
		for(i=0; i<kTransSize; ++i) {
			mTempLBuffer[i] = mOutLBuffer[mBitFwd[i]];
			mTempRBuffer[i] = mOutRBuffer[mBitFwd[i]];
		}

		VDComputeFHT(mTempLBuffer, kTransSize, mSineTab);
		VDComputeFHT(mTempRBuffer, kTransSize, mSineTab);

		// Perform convolution in Hartley domain; 0 is a special case
		mOutLBuffer[mBitRev[0]] = (mTempLBuffer[0]*mFilterLtoLs[0] + mTempRBuffer[0]*mFilterRtoLs[0]);
		mOutRBuffer[mBitRev[0]] = (mTempLBuffer[0]*mFilterLtoRs[0] + mTempRBuffer[0]*mFilterRtoRs[0]);
		for(i=1; i<kWindowSize; ++i) {
			ni = kTransSize-i;
			// If z=x*h, 2*Z(w) = X(w)(H(w)+H(-w)) + X(-w)(H(w)-H(-w)) = X(w)Hs(w) + X(-w)Hd(w)
			mOutLBuffer[mBitRev[i]]  = (mTempLBuffer[i]*mFilterLtoLs[i]   + mTempLBuffer[ni]*mFilterLtoLd[i] + mTempRBuffer[i]*mFilterRtoLs[i] + mTempRBuffer[ni]*mFilterRtoLd[i]);
			mOutLBuffer[mBitRev[ni]] = (mTempLBuffer[ni]*mFilterLtoLs[i] - mTempLBuffer[i]*mFilterLtoLd[i] + mTempRBuffer[ni]*mFilterRtoLs[i]  - mTempRBuffer[i]*mFilterRtoLd[i]);
			mOutRBuffer[mBitRev[i]]  = (mTempLBuffer[i]*mFilterLtoRs[i]   + mTempLBuffer[ni]*mFilterLtoRd[i] + mTempRBuffer[i]*mFilterRtoRs[i] + mTempRBuffer[ni]*mFilterRtoRd[i]);
			mOutRBuffer[mBitRev[ni]] = (mTempLBuffer[ni]*mFilterLtoRs[i] - mTempLBuffer[i]*mFilterLtoRd[i] + mTempRBuffer[ni]*mFilterRtoRs[i]  - mTempRBuffer[i]*mFilterRtoRd[i]);
		}

		VDComputeFHT(mOutLBuffer, kTransSize, mSineTab);
		VDComputeFHT(mOutRBuffer, kTransSize, mSineTab);

		for(i=0; i<kWindowSize; ++i) {
			mOutput[i][0] = (mOutLBuffer[i] + mOldLTail[i])*SCALE;
			mOutput[i][1] = (mOutRBuffer[i] + mOldRTail[i])*SCALE;
			mOldLTail[i] = mOutLBuffer[i+kWindowSize];
			mOldRTail[i] = mOutRBuffer[i+kWindowSize];
		}

		mInputSamplesNeeded = kWindowSize;

		return true;
	}
};

